#!/usr/bin/env python2

# Copyright (C) 2014 Red Hat
#
# This file is part of wikitcms.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>
#
# Code to get user name from fedora.cert taken from fedora-easy-karma, by
# Till Maas: https://fedorapeople.org/cgit/till/public_git/fedora-easy-karma.git

# relval manipulates release validation events.

import fedfind.exceptions

import relval.user_stats as uss
import relval.testcase_stats as tcs
import relval.report_results as rrs
import relval.size_check as rsc

import wikitcms.event as ev
import wikitcms.helpers as hl
import wikitcms.page as pg
import wikitcms.release as rl
import wikitcms.result as rs
import wikitcms.wiki as wk
from wikitcms.exceptions import NoPageError, NotFoundError, TooManyError

from mwclient import errors as mwe

import argparse
import datetime
import getpass
import os
import shutil
import smtplib
import sys
import tempfile
from collections import defaultdict, OrderedDict
from operator import attrgetter

# fedora_cert is optional. It is only used to try and guess the FAS user
# name if it is not specified on the command line.
try:
    import fedora_cert
except ImportError:
    pass

class Cli(object):

    ## UTILITY METHODS ##

    def handle_existing(self, error):
        """Convenience function for warning (but not failing) on page
        already existing, as most of the 'write page' operations want
        to do this. Use it in a try/except block, it expects to be
        passed an mwclient.errors APIError object.
        """
        error = error.args[0]
        if error == 'articleexists':
            print("Page already exists, and forced write was not requested! "
                   "Not writing.")
        else:
            print("Unhandled error writing page!")
            raise

    def get_testtypes(self, site, args):
        """Filter testtypes list from command line arguments - shared
        by multiple commands.
        """
        if args.testtype == 'all':
            testtypes = site.testtypes
        else:
            if args.testtype in site.testtypes:
                testtypes = [args.testtype]
            else:
                print("Invalid test type specified! Valid test types are:")
                for testtype in site.testtypes:
                    print(testtype)
                sys.exit("all - generates pages for all test types.")
        return testtypes

    # Comma-separated list of digits, or just one (for a list of bug IDs)
    def list_ints(self, string):
        ints = string.split(',')
        for i in ints:
            if not i.isdigit():
                err = "{} does not look like a bug ID."
                raise argparse.ArgumentTypeError(err.format(i))
        return ints

    # Just a tiny wrapper to raise the right kind of exception
    def comment_check(self, string):
        try:
            return rrs.comment_string(string)
        except ValueError as e:
            raise argparse.ArgumentTypeError(e)

    def read_creds(self):
        """Read the credentials file, if present."""
        creds = OrderedDict()
        try:
            credfile = open(os.path.expanduser('~/.fedora/credentials'), 'r')
            for line in credfile:
                linetup = line.split()
                if len(linetup) == 2:
                    creds[linetup[0].strip()] = linetup[1].strip()
                elif len(linetup) == 1:
                    creds[linetup[0].strip()] = None
                else:
                    continue
        except:
            pass
        return creds

    def guess_release(self, site, args):
        """Guess the release to operate on: uses the 'next' release so far
        as the wiki is concerned.
        """
        release = int(site.expandtemplates("{{FedoraVersionNumber|next}}"))
        print("Guessing release: {0}".format(str(release)))
        return release

    def guess_user(self, creds, interactive=True):
        """Guess username: prefer first entry in credentials file, then
        fedora.cert username, then OS username.
        """
        user = None
        if creds:
            user = creds.keys()[0]
        else:
            try:
                user = fedora_cert.read_user_cert()
            except:
                try:
                    user = getpass.getuser()
                except:
                    pass
        if not user and not interactive:
            sys.exit("No user name provided.")
        while not user:
            user = raw_input("Please enter your Fedora user name: ")
        return user

    def setup_site(self, args, creds=None):
        """Access and, if login is True, log in to either the Fedora wiki
        (test=False) or the Fedora staging wiki (test=anything else, e.g.
        True). Returns a wikitcms Wiki instance.
        """
        if not args.test:
            site = wk.Wiki(('https', 'fedoraproject.org'), '/w/')
        else:
            site = wk.Wiki(('https', 'stg.fedoraproject.org'), '/w/')
        if args.login is True:
            # password
            password = None
            try:
                password = creds[args.username]
            except:
                pass
            if not password and not args.interactive:
                    sys.exit("Could not read password from credentials "
                             "file.")
            while not password:
                password = getpass.getpass("Enter your FAS password: ",
                                           sys.stderr)
            try:
                print("Logging in with user name: {0}".format(args.username))
                site.login(args.username, password)
            except:
                sys.exit("Could not log in to wiki! Check username and "
                         "password.")
        return site

    def parse_args(self):
        """Parse command-line arguments, set up the site instance, and
        call the appropriate sub-command method.
        """
        # Help texts shared by multiple sub-commands.
        release_help = "The Fedora release to operate on (22, 23, 24...)"
        milestone_help_specific = (
            "The milestone to operate on (Alpha, Beta, Final)")
        milestone_help = (
            "The milestone to operate on (Alpha, Beta, Final, Branched, "
            "Rawhide). Must be specified for a TC/RC; for a nightly event, "
            "will be guessed if not specified")
        compose_help_specific = "The compose to operate on (TC1, RC3...)"
        compose_help = "The compose to operate on (TC1, RC3, 20150213)"
        testtype_help = "The test type to generate a page for (Base, Cloud...)"
        username_help = "Your wiki (FAS) username"
        force_help = "Force creation even if page(s) already exist"
        test_help = "Operate on the staging wiki (for testing)"
        current_help = "Do not update Test Results:Current redirect pages"
        filter_help = ("Keyword(s) to filter pages that will be processed. Can "
                      "be specified more than once, all pages matching ANY "
                      "specified keyword will be included")
        download_help = ("Only write/update the Download page, do not touch "
                        "any other page")

        # See the argparse instructions for all of this. Use of 'metavar' avoids
        # the entire ranges of valid values being shown in the help.
        parser = argparse.ArgumentParser(description=(
                "Tool for various tasks related to Fedora release validation "
                "testing. 'compose' and 'nightly' can create the result pages "
                "for TC/RC and nightly validation test events, respectively. "
                "'user-stats' and 'testcase-stats' can output statistics "
                "related to validation testing. 'report-results' lets you "
                "report test results. 'report-auto' reports test results non-"
                "interactively. See each command's help for more details."),
                                        epilog=(
                "For commands which require logging in to the wiki, you can "
                "specify a user name with --username. If you do not, relval "
                "will use the user name from ~/.fedora.cert if it exists and "
                "fedora-cert is installed, otherwise it will attempt to use "
                "your system account user name.\n\n For unattended operation, "
                "you can create a file ~/.fedora/credentials, containing one "
                "user name and one password per line, separated by a space "
                "(and readable only by your own user!). If you pass --username "
                "and specify a user listed in the credentials file, the "
                "password found in the file will be used. If a credentials "
                "file exists and --username is not passed, the first user name "
                "and password pair in the credentials file will be used. If no "
                "credentials file exists, you will be prompted for your "
                "password."))
        subparsers = parser.add_subparsers()

        parser_compose = subparsers.add_parser(
                'compose', description="Create Fedora TC/RC release validation "
                "event result pages. You must specify --milestone and "
                "--compose. This command requires login; see main help "
                "information about user name and password.")
        parser_compose.add_argument('-r', '--release', help=release_help,
                                    type=int, required=False,
                                    choices=range(22, 100), metavar="22-99")
        parser_compose.add_argument('-m', '--milestone', required=True,
                                    help=milestone_help_specific,
                                    choices=['Alpha', 'Beta', 'Final'])
        parser_compose.add_argument('-c', '--compose',
                                    help=compose_help_specific,
                                    required=True, metavar="{T,R}C1-19")
        parser_compose.add_argument('-y', '--testtype', help=testtype_help,
                                    required=False, default='all')
        parser_compose.add_argument('-u', '--username', help=username_help,
                                    required=False)
        parser_compose.add_argument('-f', '--force', help=force_help,
                                    required=False, default=True,
                                    action='store_const', const=None)
        parser_compose.add_argument('-t', '--test', help=test_help,
                                    required=False, action='store_true')
        parser_compose.add_argument('-n', '--no-current', help=current_help,
                                    required=False, action='store_true')
        parser_compose.add_argument('-w', '--download-only', required=False,
                                    help = download_help, action='store_true')
        parser_compose.set_defaults(func=self.tcrc_compose, login=True,
                                    guessrel=True, interactive=True)

        parser_nightly = subparsers.add_parser(
            'nightly', description="Create Fedora nightly release "
            "validation event result pages. If you do not specify --date, "
            "today's date will be used. If you do not specify --release, "
            "the 'next' release according to the wiki FedoraVersionNumber "
            "template will be used. If you do not specify --build, relval "
            "will try to guess the correct type based on the release "
            "chosen and the properties of the available trees, and quit "
            "if it cannot make a good enough guess. This command requires "
            "login; see main help information about user name and "
            "password.")
        parser_nightly.add_argument(
            '-c', '--compose', '-d', '--date', help="The date of the nightly "
            "compose to operate on (20141208, 20150123...)", metavar="DATE",
            type=rrs.datecheck, required=False,
            default=datetime.datetime.now().strftime('%Y%m%d'))
        parser_nightly.add_argument(
            '-r', '--release', help=release_help, type=int, required=False,
            choices=range(22, 100), metavar="22-99")
        parser_nightly.add_argument(
            '-m', '--milestone', '-b', '--build', help="The milestone of the "
            "compose (Rawhide or Branched)", required=False,
            choices=(['Rawhide', 'Branched']))
        parser_nightly.add_argument(
            '-y', '--testtype', help=testtype_help, required=False,
            default='all')
        parser_nightly.add_argument(
            '-u', '--username', help=username_help, required=False)
        parser_nightly.add_argument(
            '-f', '--force', help=force_help, required=False, default=True,
            action='store_const', const=None)
        parser_nightly.add_argument(
            '-t', '--test', help=test_help, required=False,
            action='store_true')
        parser_nightly.add_argument(
            '-n', '--no-current', help=current_help, required=False,
            action='store_true')
        parser_nightly.add_argument(
            '-i', '--if-needed', help="Only create the event if it is "
            "'needed', compared to the current test event", required=False,
            action='store_true')
        parser_nightly.add_argument(
            '-e', '--email', help="Send an announcement email for the event "
            "from the given address, if it is created", required=False)
        parser_nightly.add_argument(
            '-w', '--download-only', required=False, help=download_help,
            action='store_true')
        parser_nightly.add_argument(
            '-a', '--wait', required=False, type=int, help="Wait this number "
            "of minutes for the compose to complete before creating the "
            "event. You may want to do this to make sure the download table "
            "contains all the images, especially when running unattended/"
            "scheduled.")
        parser_nightly.set_defaults(
            func=self.nightly_compose, login=True, guessrel=True,
            interactive=True)

        parser_userstats = subparsers.add_parser(
                'user-stats', description="Output statistics regarding how "
                "many tests were run and bugs reported by each individual "
                "tester for a set of test events.")
        parser_userstats.add_argument('-r', '--release', help=release_help,
                                      type=int, required=False,
                                      choices=range(12, 100), metavar="12-99")
        parser_userstats.add_argument('-m', '--milestone', required=False,
                                      help=milestone_help_specific,
                                      choices=['Alpha', 'Beta', 'Final'])
        parser_userstats.add_argument('-t', '--test', help=test_help,
                                      required=False, action='store_true')
        parser_userstats.add_argument('-f', '--filter', metavar='KEYWORD',
                                      help=filter_help, action='append',
                                      required=False)
        parser_userstats.add_argument('-b', '--bot', help="Include results "
                                      "from automated tests ('bot results')",
                                      action='store_true', required=False)
        parser_userstats.set_defaults(func=self.user_stats, login=False,
                                      guessrel=True, interactive=True)

        parser_tcstats = subparsers.add_parser(
                'testcase-stats', description="Output a set of HTML pages "
                "containing tables providing an overview of test coverage "
                "across a series of test events.")
        parser_tcstats.add_argument('-r', '--release', help=release_help,
                                    type=int, required=False,
                                    choices=range(12, 100), metavar="12-99")
        parser_tcstats.add_argument('-m', '--milestone', required=False,
                                    help=milestone_help_specific,
                                    choices=['Alpha', 'Beta', 'Final'])
        parser_tcstats.add_argument('-o', '--out', help="The directory to write"
                                    " the pages to. WILL BE OVERWRITTEN if it "
                                    "already exists", metavar='/some/path',
                                    required=False)
        parser_tcstats.add_argument('-t', '--test', help=test_help,
                                    required=False, action='store_true')
        parser_tcstats.add_argument('-f', '--filter', metavar='KEYWORD',
                                    help=filter_help, action='append',
                                    required=False)
        parser_tcstats.add_argument('-n', '--nopostproc', help="Do not "
                                    "post-process the results to try and clean "
                                    "moved test cases and section order",
                                    required=False, action='store_true')
        parser_tcstats.set_defaults(func=self.testcase_stats, login=False,
                                    guessrel=True, interactive=True)

        parser_result = subparsers.add_parser(
                'report-results', description="Submit validation test results "
                "for an event. This command requires login; see main help "
                "information about user name and password.")
        parser_result.add_argument('-u', '--username', help=username_help,
                                    required=False)
        parser_result.add_argument('-r', '--release', help=release_help,
                                   type=int, required=False,
                                   choices=range(12, 100), metavar="12-99")
        parser_result.add_argument('-m', '--milestone', help=milestone_help,
                                   required=False, default='',
                                   choices=['Alpha', 'Beta', 'Final',
                                            'Branched', 'Rawhide'])
        parser_result.add_argument('-c', '--compose', help=compose_help,
                                   required=False, default='',
                                   metavar="{T,R}C1-19 or YYYYMMDD")
        parser_result.add_argument('-y', '--testtype', help="The test type to "
                                   "report a result for (Base, Cloud...)",
                                   required=False)
        parser_result.add_argument('-s', '--section', help="The page section "
                                   "to report the result in", required=False,
                                   default=None)
        parser_result.add_argument('-t', '--test', help=test_help,
                                   required=False, action='store_true')
        parser_result.set_defaults(func=self.report_results, login=True,
                                   guessrel=False, interactive=True)

        parser_autores = subparsers.add_parser(
                'report-auto', description="Submit validation test results for "
                "an event, non-interactively. All necessary information to "
                "uniquely identify a reporter, test instance, and result must "
                "be available to relval without interaction. This includes a "
                "username, password, a milestone if reporting for a specific "
                "TC/RC rather than a nightly or the latest event, and a test "
                "case. If --release is not specified (or is given as "
                "something other than an integer), relval will use the same "
                "release as the current validation event. If --compose is not "
                "specified, relval will use the current compose so long as "
                "it does not conflict with --release or --milestone (if either"
                " was specified). If --compose is specified as a date but no "
                "--milestone is specified, relval will first check for a "
                "Rawhide event for the specified date and specified or guessed"
                " release, then if there is not one, it will check for a "
                "Branched event. If the test case occurs more than once in "
                "the specified page, either a section name or test name or "
                "both must be provided - sufficient to identify a single test "
                "instance. The password must be available from the credentials"
                " file; the username will be determined as described in the "
                "main help. All other information must be provided via "
                "arguments. This command will quit immediately on any type of "
                "failure.")
        parser_autores.add_argument(
                '-u', '--username', help=username_help, required=False)
        parser_autores.add_argument(
                '-r', '--release', help=release_help, required=False,
                metavar="12-99")
        parser_autores.add_argument(
                '-m', '--milestone', help=milestone_help, required=False,
                choices=['', 'Alpha', 'Beta', 'Final', 'Branched', 'Rawhide'])
        parser_autores.add_argument(
                '-c', '--compose', help=compose_help,
                metavar="{T,R}C1-19 or YYYYMMDD", required=False)
        parser_autores.add_argument(
                '-y', '--testtype', help="The test type to report a result for "
                "(Base, Cloud...)", required=True)
        parser_autores.add_argument(
                '-t', '--testcase', help="The test case to report the result "
                "for. The provided string will be used as a match, so you can "
                "provide just part of the full name, but bear in mind that "
                "this argument along with --section and --testname must "
                "uniquely identify a single test instance. As a convenience, "
                "if no test case match is found, test name matches will be "
                "checked as well", required=True)
        parser_autores.add_argument(
                '-s', '--section', help="The page section to report the result "
                "in. The provided string will be used as a match, so you can "
                "provide just part of the full name, but bear in mind that "
                "this argument along with --testcase and --testname must "
                "uniquely identify a single test instance", required=False,
                default=None)
        parser_autores.add_argument(
                '-n', '--testname', help="The test name to report the result "
                "for. The 'test name' is the link text or text in the 'Test "
                "Case' cell outside of the wikilink syntax, when such text is "
                "used to distinguish between multiple occurrences of a test "
                "case in a page or page section. The provided string will be "
                "used as a match, so you can provide just part of the full "
                "name, but bear in mind that  this argument along with "
                "--testcase and --section must uniquely identify a single test "
                "instance", required=False, default=None)
        parser_autores.add_argument(
                '-e', '--environment', help="The environment to report the "
                "result for. The provided string will be used as a match, so "
                "you can provide just part of the full name, but the string "
                "you provide must match only one of the environments for the "
                "specified test instance. If there is only one 'environment' "
                "- one result column - this can be omitted", required=False,
                default='')
        parser_autores.add_argument(
                '-g', '--bugs', help="A comma-separated list of bug IDs to "
                "associate with the report", required=False,
                type=self.list_ints, metavar="123456,654321")
        parser_autores.add_argument(
                '--comment', help="A comment on the result. Must be less "
                "than 250 characters", required=False, type=self.comment_check,
                default='')
        parser_autores.add_argument(
                '--test', help=test_help, required=False,
                action='store_true')
        parser_autores.add_argument(
                '--human', help="By default results filed this way are "
                "marked as 'bot results' as it's assumed it will usually be "
                "used by automated scripts. If the result is from a manual "
                "test, pass this argument so it will not be marked as a bot "
                "result", required=False, action='store_true')
        parser_autores.add_argument(
                'status', choices=['pass', 'fail', 'warn'], help="The result "
                "status")
        parser_autores.set_defaults(
                func=self.report_auto, login=True, guessrel=False,
                interactive=False)
        parser_sizecheck = subparsers.add_parser(
            'size-check', description="Check image sizes for a compose and "
            "report the results.")
        parser_sizecheck.add_argument(
            '-u', '--username', help=username_help, required=False)
        parser_sizecheck.add_argument(
            '-r', '--release', help="The Fedora release to check", type=int,
            required=False, choices=range(1, 100), metavar="1-99")
        parser_sizecheck.add_argument(
            '-m', '--milestone', help="A milestone to check (e.g. Alpha or "
            "Beta)", choices=['Alpha', 'Beta', 'Final'], required=False)
        parser_sizecheck.add_argument(
            '-c', '--compose', help="A compose to check (e.g. TC1 or RC3)",
            required=False, metavar="{T,R}C1-19, 20150213")
        parser_sizecheck.add_argument(
            '--test', help=test_help, required=False, action='store_true')
        parser_sizecheck.set_defaults(
            func=self.size_check, login=True, guessrel=False,
            interactive=True)

        args = parser.parse_args()

        creds = self.read_creds()
        if args.login and not args.username:
            args.username = self.guess_user(creds, args.interactive)

        site = self.setup_site(args, creds)
        # guessrel is kinda ugly, but some sub-functions don't want release
        # guessing
        if args.guessrel and not args.release:
            args.release = self.guess_release(site, args)

        return (args, site)

    ## SUB-COMMAND METHODS ##

    def create_compose(self, args, site, event, eventpages):
        """Create the result pages for a given event (and, optionally,
        test type). Uses parameters passed in from the command line and
        from relevant sub-command functions, implements bits that are
        the same whatever the type of compose for which the pages are
        being created. Shared by tcrc_compose() and nightly_compose().
        """
        print("Creating pages for event {0}".format(event.version))
        if args.download_only:
            print("Just handling Download page...")
            eventpages = (pg.DownloadPage(site, event),)
        else:
            eventpages.append(pg.SummaryPage(site, event))
            eventpages.append(pg.DownloadPage(site, event))
        for page in eventpages:
            # Only do the category fixup if page creation succeeds.
            try:
                print("Writing page: " + page.name)
                page.write(createonly=args.force)
            except mwe.APIError as error:
                self.handle_existing(error)
            if args.download_only:
                # Short-circuit outta here!
                sys.exit()
            # Update current redirect pages, if asked to.
            if not args.no_current:
                print("Updating Current redirect page...")
                page.update_current()

        # Update the template that defines the 'current' compose (if asked)
        if not args.no_current:
            try:
                event.update_current()
            except:
                print("Unhandled error updating current template!")
                raise

        # Make sure the category hierarchy is all in place.
        print("Creating event category pages if necessary (or forced)...")
        for page in [event.category_page, event.parent_category_page]:
            try:
                page.write(createonly=args.force)
            except mwe.APIError as error:
                self.handle_existing(error)
            except:
                print("Unhandled error writing category page!")
                raise

    def tcrc_compose(self, args, site):
        """Create the result pages for a TC/RC compose event, by
        getting the TC/RC compose type specific bits and passing them
        along to create_compose().
        """
        try:
            event = site.get_validation_event(
                release=args.release, milestone=args.milestone,
                compose=args.compose)
        except ValueError:
            sys.exit("Compose must be a TC/RC identifier (TC1, RC3...)")
        testtypes = self.get_testtypes(site, args)
        eventpages = [pg.ComposePage(site, args.release, t, args.milestone,
                                     args.compose)
                      for t in testtypes]
        self.create_compose(args, site, event, eventpages)

    def nightly_compose(self, args, site):
        """Create the result pages for a nightly compose test event,
        by getting the nightly compose type specific bits and passing
        them along to create_compose().
        """
        curr = None
        nocompose = True
        if args.milestone:
            tree = args.milestone
            event = ev.NightlyEvent(
                site, args.release, args.milestone, args.compose)
            if event.compose_exists:
                nocompose = False
        else:
            # We weren't told which tree (Rawhide or Branched) to use, so we're
            # going to have to guess. The following encapsulates some specific
            # knowledge of how the Fedora release cycle works, and may need
            # changing if it changes. The point of this is to make it viable to
            # have 'relval nightly --if-needed' running daily on a cron job
            # somewhere, and have it only create an event when it's sure it's
            # doing the right thing. For this guessing to work, the nightly
            # compose(s) must at least have *started* when the command is run.
            curr = site.current_compose
            if curr['release'] and args.release == int(curr['release']) + 1:
                # First nightly event after a release should always be Rawhide
                event = ev.NightlyEvent(site, args.release, 'Rawhide',
                                        args.compose)
                if event.compose_exists:
                    tree = 'Rawhide'
                    nocompose = False
            elif (curr['milestone'] and curr['milestone'] == 'Branched' and
                  curr['release'] and args.release == int(curr['release'])):
                # Branched should never be directly succeeded by Rawhide
                event = ev.NightlyEvent(site, args.release, 'Branched',
                                        args.compose)
                if event.compose_exists:
                    tree = 'Branched'
                    nocompose = False
            elif (curr['milestone'] and curr['milestone'] == 'Rawhide' and
                  curr['release'] and args.release == int(curr['release'])):
                # if current is Rawhide, prefer Branched if it exists (this
                # covers the branch point), otherwise another Rawhide is OK
                event = ev.NightlyEvent(site, args.release, 'Branched',
                                        args.compose)
                if event.compose_exists:
                    tree = 'Branched'
                    nocompose = False
                else:
                    event = ev.NightlyEvent(site, args.release, 'Rawhide',
                                            args.compose)
                    if event.compose_exists:
                        tree = 'Rawhide'
                        nocompose = False
            else:
                sys.exit("You didn't specify --milestone, and relval "
                         "can't figure out where we are in the release cycle. "
                         "It is not safe for relval to guess how to proceed. "
                         "Please specify a milestone.")

        if nocompose:
            sys.exit("Could not find a compose for the specified or detected "
                     "milestone and date {0}!".format(event.version))

        if args.wait:
            try:
                event.ff_release.wait(args.wait)
            except fedfind.exceptions.WaitError:
                sys.exit("Timed out waiting for compose to complete!")

        if not event.has_bootiso:
            sys.exit("Specified or detected compose {0} has no boot.iso! It is "
                     "not testable.".format(event.version))

        if args.if_needed:
            print("Checking if new nightly test event needed for {0}".format(
                    event.version))
            newrelease = False
            curr = site.current_compose
            if curr['release'] and args.release > int(curr['release']):
                newrelease = True
                print("Requested release is newer than release of current test"
                       " event! Proceeding.")
            else:
                try:
                    currdate = datetime.datetime.strptime(curr['date'],
                                                          '%Y%m%d')
                except:
                    if curr['compose']:
                        sys.exit("Current test event is for a TC/RC of the same"
                                 " release! --needed will not succeed a TC/RC "
                                 "test event for the same release.")
                    else:
                        sys.exit("Cannot determine date of current nightly test"
                                 " event!")
                newdate = datetime.datetime.strptime(args.compose, '%Y%m%d')
                delta = newdate - currdate
                if delta.days < 3:
                    sys.exit("Less than three days since the last nightly test "
                             "event.")
                if delta.days > 14:
                    print("More than two weeks since last nightly test event!")
                else:
                    print("Comparing package versions...")
                    packages = ['anaconda', 'python-blivet', 'pyparted',
                                'lorax', 'pungi', 'parted', 'pykickstart']
                    currevent = ev.NightlyEvent(site, curr['release'],
                                                curr['milestone'], curr['date'])
                    try:
                        currpacks = currevent.get_package_versions(packages)
                        newpacks = event.get_package_versions(packages)
                    except:
                        sys.exit("Package version check failed!")
                    if not currpacks:
                        sys.exit("Could not do package version check on current "
                                 "compose!")
                    if not newpacks:
                        sys.exit("Could not do package version check on new "
                                 "compose!")
                    for key in currpacks.keys():
                        if not currpacks[key] or not newpacks[key]:
                            sys.exit("Could not find versions for all "
                                     "significant packages in both composes!")
                    if currpacks == newpacks:
                        sys.exit("No significant package changes since last "
                               " nightly test event.")
                    else:
                        print("Significant package changes since last nightly "
                               "test event!")
                        diff = ''
                        for key in currpacks.keys():
                            if currpacks[key] != newpacks[key]:
                                diff += ("{0} - {1}: {2}, {3}: {4}\n".format(
                                        key, curr['date'], currpacks[key],
                                        args.compose, newpacks[key]))
                        print(diff)

        testtypes = self.get_testtypes(site, args)
        eventpages = [pg.NightlyPage(site, args.release, t, tree, args.compose)
                      for t in testtypes]
        self.create_compose(args, site, event, eventpages)

        if args.email:
            print("Sending announcement email...")
            if args.test:
                dest = args.email
            else:
                dest = "test-announce@lists.fedoraproject.org"
            urltmpl = "https://fedoraproject.org/wiki/{0}"
            summurl = urltmpl.format(event.summary_page.name.replace(' ', '_'))
            pageurls = [urltmpl.format(p.name.replace(' ', '_')) for p in
                        eventpages if u'Summary' not in p.name]
            try:
                difftext = '\nNotable package version changes:\n{0}'.format(
                        diff)
            except:
                difftext = ''
            tmpl = ("""From: {0}
To: {1}
Subject: Fedora {2} nightly compose nominated for testing

Announcing the creation of a new nightly release validation test event
for Fedora {2}. Please help run some tests for this
nightly compose if you have time. For more information on nightly
release validation testing, see:
https://fedoraproject.org/wiki/QA:Release_validation_test_plan
{3}
Test coverage information for the current release can be seen at:
https://www.happyassassin.net/testcase_stats/{4}

You can see all results, find testing instructions and image download
locations, and enter results on the Summary page:

{5}

The individual test result pages are:

{6}

Thank you for testing!
-- 
Mail generated by relval: https://www.happyassassin.net/wikitcms/
On behalf of: {7}
""")
            msg = tmpl.format(args.email, dest, event.version, difftext,
                              event.release, summurl, '\n'.join(pageurls),
                              args.username)
            server = smtplib.SMTP('localhost')
            server.sendmail(args.email, [dest], msg)
            print("Announcement mail sent from {0} to {1}!".format(args.email,
                                                                   dest))

    def user_stats(self, args, site):
        """Print HTML-formatted statistics on the number of results
        reported by each tester for a given set of result pages.
        """
        release = rl.Release(args.release, site)
        pages = release.milestone_pages(args.milestone)
        if args.filter:
            pages = [p for p in pages for f in args.filter if f in p.name]
        pages = sorted(pages, key=lambda page: page.sorttuple)
        text = []
        # dump the results tables from each page into a single big string
        for page in pages:
            sys.stderr.write("Processing page: {}\n".format(page.name))
            text.append(page.results_wikitext)
        text = '\n'.join(text)

        # find results with the statuses we care about that are not transferred
        # from previous runs
        results = rs.find_results(text, statuses=['pass', 'fail', 'warn'],
                                    transferred=False, bot=args.bot)

        # make two defaultdicts for number of reports and cited bugs per unique
        # user. defaultdict(int) sets the value when adding a key to int(0).
        reports = defaultdict(int)
        bugs = defaultdict(set)
        for result in results:
            if result.user:
                reports[result.user.lower()] += 1
            if result.bugs:
                bugs[result.user].update(result.bugs)

        uss.print_stats(args.release, reports, bugs)

    def testcase_stats(self, args, site):
        """Generate HTML-formatted statistics on the test coverage for
        a given release - produces summary pages for each test type
        containing tables with a row for each test indicating the level
        of coverage at each compose, and detail pages for each test
        with detailed statistics on the test results and bugs for that
        test for each compose.
        """
        outdir = tempfile.mkdtemp(prefix = 'tc_report_')
        release = rl.Release(args.release, site)
        relpages = release.milestone_pages(args.milestone)
        if args.filter:
            relpages = [p for p in relpages for f in args.filter if f in
                        p.name]

        allpages = defaultdict(list)
        allcomposes = set()
        for page in relpages:
            allpages[page.testtype].append(page)
            allcomposes.add(page.shortver)

        # turn the composes set into a list and sort it, using the Fedora
        # release ordering sort function wikitcms.helpers provides.
        allcomposes = list(allcomposes)
        allcomposes.sort(key=hl.fedora_release_sort)

        # for each test type, sort the pages into the correct Fedora release
        # order, then extract the result rows from each page.
        #
        # 'tests' is an ordered dict of Test() objects, keyed with a tuple
        # identifying 'unique tests': the test 'name', its testcase page name,
        # and the name of the page section in which it appears. It is ordered
        # so that tests appear in the order they are encountered - when the
        # HTML output is rendered, the ordering of tests will be more or less
        # the same as on the actual result wiki pages. tcs.post_process()
        # cleans this up some more once all result pages have been parsed. See
        # its docs for details.
        #
        # A Test() object is generated for each result row from each result
        # page for the test type, and added to the tests dict based on those
        # three attributes (as the second field of a tuple, the first field
        # identifying the compose from which it comes). Result rows for which
        # all three attributes match are grouped together - they're considered
        # to be results for the same test in different composes.
        for (testtype, pages) in allpages.iteritems():
            tests = OrderedDict()
            pages.sort(key=attrgetter('sorttuple'))
            for page in pages:
                print("Processing: " + page.name)
                resultrows = page.get_resultrows(
                        statuses=['pass', 'warn', 'fail'], transferred=False)
                for row in resultrows:
                    tests.setdefault((row.name, row.testcase, row.section),
                                     tcs.Test())
                    tests[(row.name, row.testcase, row.section)].update(
                            page.shortver, row)

            if args.nopostproc:
                print("nopostproc passed! Skipping post-process.")
            else:
                print("Post-processing results...")
                tests = tcs.post_process(tests, allcomposes)
            tcs.print_results_html(testtype, pages, allcomposes, tests, outdir)

        tcs.print_summary_html(outdir, allpages)
        if args.out:
            try:
                shutil.rmtree(args.out)
            except:
                pass
            try:
                shutil.copytree(outdir, args.out)
                print("All output moved to: " + args.out)
                print("Index page at: " + os.path.join(args.out, "index.html"))
            except:
                print("Could not write to: {}! Output is at: {}".format(
                        args.out, outdir))
        else:
            print "HTML Output: " + os.path.join(outdir, "index.html")
            print outdir

    def report_results(self, args, site):
        """Report some validation testing results."""
        rrs.report_results(site, args.username, args.release, args.milestone,
                           args.compose, args.testtype)

    def report_auto(self, args, site):
        """Non-interactive reporting function; requires all parameters
        necessary to identify the result to be reported to be passed,
        exits on any failure.
        """
        try:
            print("Will report {} for:".format(args.status))
            print("Test case: {}".format(args.testcase))
            if args.testname:
                 print("Test name: {}".format(args.testname))
            print("Environment: {0}".format(args.environment))
            if args.section:
                print("Section: {}".format(args.section))
            print("Page: {}".format(' '.join((args.release, args.milestone,
                                               args.compose))))
            restup = (args.testtype, args.release, args.milestone,
                      args.compose, args.testcase, args.section, args.testname,
                      args.environment, args.status, args.username, args.bugs,
                      args.comment, not args.no_bot)
            print(restup)
            (insuff, dupe) = site.report_validation_results((restup,))
            if dupe:
                sys.exit("This user has already filed a report for this test "
                         "and environment! Will not file multiple reports.")
            if insuff:
                sys.exit("Insufficient information provided to identify a "
                         "page, test instance, and result!")
            print("Result submission succeeded!")
            sys.exit(0)
        except:
            raise

    def size_check(self, args, site):
        try:
            event = site.get_validation_event(
                args.release, args.milestone, args.compose)
            ffrel = event.ff_release
            print("Checking sizes for: {0}").format(event.version)
        except ValueError:
            sys.exit("Could not find validation test event!")
        results = rsc.size_check(ffrel)
        if not results:
            sys.exit("No images found for event!")
        reslist = list()
        for (env, result) in results.items():
            if result == 'pass':
                restup = wk.ResTuple(
                    'Installation', args.release, args.milestone, args.compose,
                    'ISO_Size', '', '', env, result, args.username, '', '',
                    True)
            else:
                # Fail - construct a comment with details
                comms = ("{0}, size {1}, max {2}".format(fail[0].desc,
                         fail[1], fail[2]) for fail in result)
                comment = "<ref>{0}</ref>".format('</ref><ref>'.join(comms))
                # Mark this as a bot result as it's an automated test.
                restup = wk.ResTuple(
                    'Installation', args.release, args.milestone, args.compose,
                    'ISO_Size', '', '', env, 'fail', args.username, '',
                    comment, True)
            reslist.append(restup)
        (insuffs, dupes) = site.report_validation_results(reslist)
        if dupes:
            for dupe in dupes:
                print("This user already reported a result for {0}! Will not "
                      "file duplicate report.".format(dupe.env))
        if insuffs and len(insuffs) == len(results):
            sys.exit("Could not find result page or test case!")
        if insuffs:
            for insuff in insuffs:
                print("Could not find environment {0}!".format(insuff.env))
        sys.exit(0)

    ## MAIN LOOP BITS ##

    def run(self):
        args, site = self.parse_args()
        args.func(args, site)

## END OF Cli CLASS ##

def main():
    '''Main loop'''
    try:
        cli = Cli()
        cli.run()
    except KeyboardInterrupt:
        print 'Interrupted, exiting...'
        sys.exit(1)

if __name__ == '__main__':
    main()
