# Copyright (C) 2014 Red Hat
#
# This file is part of relval.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
# Adam Williamson <awilliam@redhat.com>
#
# relval interacts with Fedora release validation events. This file contains
# functions that help implement the 'report-results' sub-command of relval,
# which is a simple utility to report release validation results.

from wikitcms import event as ev
from wikitcms import page as pg
from wikitcms import wiki as wk
from wikitcms import result as rs

import sys
import re
import datetime

white = re.compile('\s')
compose_patt = re.compile('[TR]C\d+\.?\d*')
tag_patt = re.compile('<.*?>')

# date checker, note relval itself also uses this
def datecheck(string):
    try:
        date = datetime.datetime.strptime(string, '%Y%m%d')
    except:
        raise ValueError('Date must be like 20141208')
    return string

# Little generic function for only allowing a string to be a certain length.
# If it's too long, returns a 2-tuple of the actual length and the max length.
def max_string(string, maxlen=250):
    string = string.strip()
    if maxlen and len(string) > maxlen:
        raise ValueError(len(string), maxlen)
    return string

# String wrapped in ref tags. By default, max 250 characters.
# relval itself also uses this, in report-auto
def comment_string(string, maxlen=250):
    try:
        string = max_string(string, maxlen)
    except ValueError as e:
        err = "Comment is too long: {} characters, max is {}."
        raise ValueError(err.format(e[0], e[1]))
    # Don't produce an empty <ref></ref> if there's no comment
    if not string:
        return string
    return "<ref>{}</ref>".format(string)

# mini functions for prompting the user for choices

def get_confirm():
    confirm = raw_input("Confirm y/n? ")
    while confirm.lower() not in {'y', 'n'}:
        print("Invalid response.")
        confirm = get_confirm()
    return confirm.lower()

def get_release():
    release = raw_input("Which Fedora release would you like to report a "
                            "result for? ")
    try:
        release = int(release)
    except:
        pass
    return release

def get_version_type():
    print("(C)ompose - test compose or release candidate")
    print("(N)ightly build")
    type = raw_input("Which would you like to report a result for? ")
    if type.lower() == 'c' or type.lower() == 'n':
        return type.lower()
    else:
        return None

def get_milestone():
    print("(A)lpha")
    print("(B)eta")
    print("(F)inal")
    m = raw_input("Which milestone would you like to report a milestone for? ")
    if m.lower() == 'a':
        return 'Alpha'
    elif m.lower() == 'b':
        return 'Beta'
    elif m.lower() == 'f':
        return 'Final'
    else:
        return None

def get_compose():
    compose = raw_input("Which compose would you like to report a result for? ")
    match = compose_patt.match(compose)
    if match:
        return match.group()
    return None

def get_date():
    date = raw_input("What date would you like to report a result for? ")
    try:
        datecheck(date)
    except:
        return None
    return date

def get_testtype():
    t = raw_input("Which test type would you like to report a result for? ")
    try:
        return int(t)
    except:
        return None

def get_section():
    section = raw_input("Which section would you like to report a result in? ")
    try:
        return int(section)
    except:
        return None

def get_test():
    test = raw_input("Which test would you like to report a result for? ")
    try:
        return int(test)
    except:
        return None

def get_env():
    env = raw_input("Which environment would you like to report a result for? ")
    try:
        return int(env)
    except:
        return None

def get_result(username):
    status = raw_input("Would you like to report a (P)ass, (F)ail or (W)arn? ")
    while status.lower() not in {'p', 'f', 'w'}:
        status = raw_input("Incorrect choice. Please choose (P)ass, (F)ail or "
                           "(W)arn. ")
    if status.lower() == 'p':
        status = 'pass'
    if status.lower() == 'f':
        status = 'fail'
    if status.lower() == 'w':
        status = 'warn'
    bugs = raw_input("Enter the IDs of any bugs associated with this report, "
                     "separated by commas. ")
    if bugs == '':
        bugs = None
    else:
        bugs = white.sub('', bugs)
        bugs = bugs.split(',')
        if bugs == '':
            bugs = None
    comment = ''
    comm = raw_input("Enter a comment if you like. If not, just hit Enter.\n"
                        "Maximum length is 250 characters.\nPlease don't leave "
                        "a comment unless it's really necessary.\nComment: ")
    while comm:
        try:
            comment = comment_string(comm, 250)
            comm = ''
        except ValueError:
            comm = raw_input("Comment was too long: {} characters. Please make "
                             "it 250 characters or fewer.\nComment: ".format(
                                    len(comm.strip())))
    return rs.Result(status, username, bugs, comment)

def get_cont():
    cont = raw_input("Report another result for this (s)ection, or (p)age, "
                     "or this (c)ompose, or (q)uit? ")
    while cont.lower() not in {'s', 'p', 'c', 'q'}:
        print("Invalid choice. Please choose again.")
        cont = get_cont()
    return cont

# main function
def report_results(wiki, username, release=None, milestone='', compose='',
                   testtype=None, section=None, test=None, auto=True,
                   checkver=True):
    if not username:
        sys.exit("ERROR: No valid user name!")
    # these are global so they can be cached between runs
    global respages
    global page
    global sections
    version = ''
    version_set = False
    version_type = ''

    # this means 'try and detect the current compose, unless we already
    # have one fully specified or we've been told not to'
    if not (release and compose) and auto is True:
        if release or compose:
            print("When specifying the compose version by parameters, you "
                  "must pass at least a release and a compose. Your "
                  "version-related parameters will not be used.")
        print("Detecting current compose version...")
        curr = wiki.current_compose
        release = int(curr['release'])
        milestone = curr['milestone']
        # this is a handy way to get whichever is set, and if for some
        # crazy reason both are set it'll prefer 'T/RCx' to 'YYYYMMDD',
        # which we probably want.
        compose = max(curr['compose'], curr['date'])

    if release and compose:
        version_set = True
        if milestone:
            dispver = "{0} {1} {2}".format(release, milestone, compose)
        else:
            dispver = "{0} {1}".format(release, compose)
        if checkver:
            print("Will report results for compose: {0}".format(dispver))
            confirm = get_confirm()
            if confirm == 'y':
                pass
            else:
                report_results(wiki, username, auto=False)

    if release not in range(21,100):
        release = get_release()
    while release not in range(21,100):
        print("Invalid release. Please input a valid Fedora release number. ")
        release = get_release()

    while not version_set:
        if not version_type:
            version_type = get_version_type()
        while not version_type:
            print("Invalid choice. Please choose (c)ompose or (n)ightly. ")
            version_type = get_version_type()

        if version_type == 'c':
            if milestone not in {'Alpha', 'Beta', 'Final'}:
                milestone = get_milestone()
            while not milestone:
                print("Invalid choice. Please choose (A)lpha, (B)eta, or "
                      "(F)inal. ")
                milestone = get_milestone()
            if not compose:
                compose = get_compose()
            while not compose:
                print("Invalid compose name. Examples: RC1, TC3, RC4.1 ")
                compose = get_compose()
            version_set = True
        else:
            if not compose:
                compose = get_date()
            while not compose:
                print("Invalid date. Date must be like 20141208 ")
                compose = get_date()
            version_set = True

    try:
        # This list isn't going to change within a single compose, so we
        # only bother getting it once.
        respages
    except:
        try:
            event = wiki.get_validation_event(
                release=release, milestone=milestone, compose=compose)
        except ValueError as err:
            print(err[0])
            report_results(wiki, username)

        respages = event.result_pages
        if not respages:
            print("Error! You probably chose a compose that doesn't exist.")
            report_results(wiki, username)

    try:
        page.version
    except:
        page = None
        if testtype:
            try:
                page = [p for p in respages if p.testtype == testtype][0]
            except:
                print("Error! Specified test type does not exist.")
        if not page:
            for i, page in enumerate(respages):
                print(str(i) + ")    " + page.testtype)
            t = get_testtype()
            while t not in range(0, len(respages)):
                print("Incorrect choice. Please choose one of the test types.")
                t = get_testtype()
            page = respages[t]

    tests = page.get_resultrows(
            statuses=['pass', 'warn', 'fail', 'inprogress'])

    if not section:
        try:
            # This serves two purposes: if we somehow wound up with a dead
            # page, we can try again...
            if len(sections) == 0:
                print("Error! No results found for this page.")
                report_results(wiki, username)
        except:
            # ...and if we hit an exception, it means we need to get sections.
            sections = page.results_sections
            # Filter out sections with no tests in them. wikitcms should
            # probably do this but it's actually not trivial for $REASONS
            testsecs = {t.secid for t in tests}
            sections = [s for s in sections if s['index'] in testsecs]
        if len(sections) == 1:
            section = sections[0]
        else:
            for i, section in enumerate(sections):
                print(str(i) + ")    " + tag_patt.sub('', section['line']))
            s = get_section()
            while s not in range(0, len(sections)):
                print("Incorrect choice. Please choose one of the sections.")
                s = get_section()
            section = sections[s]

    # Find the tests for the section the user chose.
    tests = [t for t in tests if t.secid == section['index']]
    if len(tests) == 1:
        t = 0
    else:
        for i, test in enumerate(tests):
            if test.name != test.testcase:
                print(str(i) + ")    " + test.testcase + " " + test.name)
            else:
                print(str(i) + ")    " + test.testcase)
        t = get_test()
        while t not in range(0, len(tests)):
            print("Incorrect choice. Please choose one of the tests.")
            t = get_test()
    test = tests[t]

    envs = test.results.keys()
    if len(envs) == 1:
        env = envs[0]
    elif len(envs) > 1:
        for i, env in enumerate(envs):
            print(str(i) + ")    " + env)
        e = get_env()
        while e not in range(0, len(envs)):
            print("Incorrect choice. Please choose an environment.")
            e = get_env()
        env = envs[e]

    results = test.results[env]
    if results:
        print("Current results for this test:")
        for result in results:
            print(str(result))
        myres = [r for r in results if r.user == username]
        if myres:
            print("You have already reported result(s) for this test! Do you "
                   "still want to report this result?")
            confirm = get_confirm()
            if confirm == 'y':
                pass
            else:
                report_results(wiki, username, release, milestone, compose)
    res = get_result(username)
    print("Will report {} for test {} in section {} of {}".format(
            res, test.testcase, section['line'], page.name))
    try:
        page.add_result(res, test, env)
        print("Succeeded!")
    except:
        raise

    cont = get_cont()
    if cont == 'q':
        sys.exit()
    else:
        test = None
    if cont == 'p' or cont == 'c':
        section = None
    if cont == 'c':
        testtype = None
        page = None
        sections = None
    report_results(wiki, username, release, milestone, compose, testtype,
                   section, checkver=False)
