# Copyright (C) 2015 Red Hat
#
# This file is part of relval.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>
#
# relval interacts with Fedora release validation events. This file contains
# functions that help implement the 'size_check' sub-command of relval,
# which checks the sizes of images and reports the results.

import fedfind.release
import requests

# https://fedoraproject.org/wiki/QA:Testcase_Mediakit_ISO_Size
# https://fedoraproject.org/wiki/Releases/22/Spins
MAXSIZES = dict(
    netinst = 734003200,
    boot = 734003200,
    cloud = 734003200,
    # This is the same as cloud_atomic_canned, it's a DVD-size image
    cloud_atomic_boot = 4700000000,
    kde_live = 1400000000,
    workstation_live = 2000000000,
    design_suite_live = 2000000000,
    lxde_live = 734003200,
    mate_live = 1500000000,
    security_live = 1000000000,
    soas_live = 734003200,
    xfce_live = 2000000000
)

STDTYPES = ('netinst', 'boot', 'dvd', 'live')

# These are the 'environments' (i.e. columns) for the test case we're
# testing here, in the Installation results page. The value for each
# key is an iterable of fedfind queries to find the images that should
# be tested to decide if that env 'passes' or 'fails'.
ENVS = dict(
    Server = (fedfind.release.Query('flavor', ('server',)),
              fedfind.release.Query('imagetype', STDTYPES)),
    Workstation = (fedfind.release.Query('flavor', ('workstation',)),
                   fedfind.release.Query('imagetype', STDTYPES)),
    Cloud = (fedfind.release.Query('flavor', ('cloud',)),
             fedfind.release.Query('imagetype', STDTYPES)),
    Spins = (fedfind.release.Query('flavor', ('',)),
             fedfind.release.Query('imagetype', ('live',)))
)

def size_check(ffrel):
    """Do the thing! Check the things! CHECK ALL THE THINGS"""
    results = dict()
    for (env, queries) in ENVS.items():
        print("Checking {0}...".format(env))
        fails = list()
        imgs = ffrel.find_images(queries)
        if not imgs:
            print("No images found!")
            continue

        for img in imgs:
            maxsize = 4700000000
            try:
                maxsize = MAXSIZES[img.imagetype]
            except KeyError:
                pass
            try:
                maxsize = MAXSIZES[img.payload]
            except KeyError:
                pass
            try:
                maxsize = MAXSIZES['{0}_{1}'.format(img.payload, img.imagetype)]
            except KeyError:
                pass
            resp = requests.head(img.url)
            imgsize = int(resp.headers['content-length'])
            if imgsize > maxsize:
                fails.append((img, imgsize, maxsize))
        if fails:
            results[env] = fails
            for fail in fails:
                print("FAIL: {0}, size {1}, max {2}".format(fail[0].filename,
                      fail[1], fail[2]))
        else:
            results[env] = 'pass'
            print("PASS!")
    return results
